Imports System.IO

Public Class frmAsyncStreaming
    Private Const _Output As String = "Output"
    Private Const _Src As String = "Output\TestSource.Data"
    Private Const _Dest As String = "Output\TestDest"

    Private Sub CreateTestSource()
        Me.RichTextBox1.AppendText("erstelle TestSource - bitte warten" & _
            Lf & Lf)

        ' Der Using-Block ist zeitaufwndig und friert das Gui ein. Vorher 
        '  die Neuzeichnung der Richtextbox explizit erzwingen
        Me.RichTextBox1.Refresh()

        If Not Directory.Exists(_Output) Then Directory.CreateDirectory(_Output)
        ' Output-Ordner im Browser anzeigen
        Diagnostics.Process.Start(_Output)
        Using _
            Reader As New FileStream(Application.ExecutablePath, _
            FileMode.Open, FileAccess.Read), _
            Writer As New FileStream(_Src, FileMode.Create)

            ' Groe Datei generieren ( > 30MB)
            For I As Integer = 0 To 999
                Reader.Position = 0
                StreamX.WriteTo(Reader, Writer)
            Next
        End Using
    End Sub

    Private Sub Button_Click(ByVal sender As Object, ByVal e As EventArgs) _
        Handles btStreamBegin.Click, btDelegatBegin.Click, _
        btExplicitThread.Click, btCopySynchron.Click, btBackgroundworker.Click

        Static ID As Integer
        ID += 1
        Dim Dest As String = _Dest & ID
        Dim Cause As String = ""
        Dim sFormat As String = "{0,-14:HH:mm:ss.FFF} starte {1} -> {2}"

        If Not File.Exists(_Src) Then CreateTestSource()
            Select Case True
                Case sender Is btStreamBegin
                    Dest &= ".StreamBeginRead"
                    Cause = "Stream.BeginRead()"
                    CopyFileMS(_Src, Dest)

                Case sender Is btDelegatBegin
                    ' Mein Favorit - keine zustzlichen Methoden erforderlich 
                    '  (bzw. die zustzlichen Methoden sind allgemein 
                    '  wiederverwendbar)
                    Cause = "Delegate-Invoking"
                    Dest &= ".Invoking"
                    CrossThread.RunAsync(AddressOf CopyFile, Cause, _Src, Dest)

                Case sender Is btExplicitThread
                    Cause = "explicit Threading"
                    Dest &= ".Threading"
                    Dim Tr As New Threading.Thread( _
                        AddressOf ExplicitThreadStart)
                    Tr.Start(New String() {Cause, _Src, Dest})

                Case sender Is btBackgroundworker
                    Dest &= ".Backgroundworker"

                    ' BackgroundWorker hat die Eigenart, da er nur einen Job 
                    '  erledigen kann 
                    If BackgroundWorker1.IsBusy Then
                        Cause = "BackgroundWorker1.IsBusy = True " & _
                            "- Kann grad nicht"
                        Exit Select
                    End If
                    Cause = "Backgroundworker"
                    BackgroundWorker1.RunWorkerAsync(New String() _
                        {Cause, _Src, Dest})

                Case sender Is btCopySynchron
                    Cause = "synchrones kopieren"
                    Dest &= ".Synchron"
                    WriteLine(String.Format(sFormat, Now, _
                        Cause, Path.GetFileName(Dest)))
                    CopyFile(Cause, _Src, Dest)
                    Return
        End Select
        ' Diese Meldung wird **vor** den Erfolgsmeldungen aus den 
        ' Nebenthreads angezeigt.
        WriteLine(String.Format(sFormat, Now, Cause, Path.GetFileName(Dest)))
    End Sub

    Private Sub ExplicitThreadStart(ByVal Obj As Object)
        Dim Args As String() = DirectCast(Obj, String())
        CopyFile(Args(0), Args(1), Args(2))
    End Sub

    Private Sub BackgroundWorker1_DoWork(ByVal sender As Object, _
        ByVal e As System.ComponentModel.DoWorkEventArgs) _
        Handles BackgroundWorker1.DoWork

        ' Der Vollstndigkeit halber hier auch die Verwendung des 
        ' BackgroundWorkers. Dieser bietet noch zustzlich das Event 
        ' _RunWorkerCompleted, welches hier aber nicht genutzt wird, da 
        ' CopyFile schon einen threadsicheren Mechanismus implementiert, um die 
        '  Erfolgsmeldung anzuzeigen
        Dim Args As String() = DirectCast(e.Argument, String())
        CopyFile(Args(0), Args(1), Args(2))
    End Sub

    Private Sub CopyFile(ByVal Cause As String, ByVal Src As String, _
        ByVal Dest As String)

        ' Nur das Threading mit Stream.BeginRead kann diese Methode in dieser 
        '  Form nicht verwenden
        Using Reader As New FileStream(Src, _
            FileMode.Open, FileAccess.Read, FileShare.Read), _
            Writer As New FileStream(Dest, FileMode.Create)

            StreamX.WriteTo(Reader, Writer)
        End Using
        WriteLine(String.Format("{0,-14:HH:mm:ss.FFF} beende {1} ({2})", _
            Now, Cause, Path.GetFileName(Dest)))
    End Sub

    Private Sub WriteLine(ByVal Text As String)
        CrossThread.RunGui(AddressOf Me.RichTextBox1.AppendText, Text & Lf)
    End Sub

#Region "Asynchroner Kopiervorgang in der von MS vorgesehenen Umstndlichkeit"

    Private Structure AsyncInfo
        Public Reader As Stream
        Public Buffer As Byte()
        Public WriterName As String

        Public Sub New(ByVal Reader As Stream, _
            ByVal Buffer As Byte(), ByVal WriterName As String)

            Me.Reader = Reader
            Me.Buffer = Buffer
            Me.WriterName = WriterName
        End Sub
    End Structure

    Private Sub CopyFileMS(ByVal Src As String, ByVal Dest As String)
        Dim Buf(Short.MaxValue - 1) As Byte
        Dim Reader As New FileStream( _
            Src, FileMode.Open, FileAccess.Read, FileShare.Read, _
            Buf.Length, FileOptions.Asynchronous)
        Reader.BeginRead( _
            Buf, 0, Buf.Length, AddressOf Callback, _
            New AsyncInfo(Reader, Buf, Dest))
    End Sub

    Private Sub Callback(ByVal Ar As IAsyncResult)
        With DirectCast(Ar.AsyncState, AsyncInfo)
            ' Suboptimales Design: Der Read-Stream muss in einer anderen 
            '  Methode und in einem anderen Thread disposed werden, als in dem 
            '  er erstellt wurde

            Using Reader As Stream = _
                .Reader, Writer As New FileStream(.WriterName, FileMode.Create)

                Dim Portion As Integer = Reader.EndRead(Ar)
                Writer.Write(.Buffer, 0, Portion)
                While Portion = .Buffer.Length
                    Portion = Reader.Read(.Buffer, 0, .Buffer.Length)
                    Writer.Write(.Buffer, 0, Portion)
                End While
            End Using
        End With
        WriteLine(String.Format("{0,-14:HH:mm:ss.FFF} beende " & _
            "Stream.EndRead()", Now))
    End Sub

    'Asynchroner Kopiervorgang in der von MS vorgesehenen Umstndlichkeit
#End Region

    Private Sub frmAsyncStreaming_FormClosed(ByVal sender As Object, _
        ByVal e As FormClosedEventArgs) Handles MyBase.FormClosed

        ' Das schliet kurioserweise den Browser (ich glaub, der strzt ab)
        Directory.Delete(_Output, True)
    End Sub

End Class
